'use strict';

var clone = require('@turf/clone');
var helpers = require('@turf/helpers');
var meta = require('@turf/meta');
var Flatbush = require('flatbush');
var booleanIntersects = require('@turf/boolean-intersects');
var booleanContains = require('@turf/boolean-contains');
var booleanWithin = require('@turf/boolean-within');
var booleanCrosses = require('@turf/boolean-crosses');
var booleanOverlap = require('@turf/boolean-overlap');
var bbox = require('@turf/bbox');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var clone__default = /*#__PURE__*/_interopDefaultLegacy(clone);
var Flatbush__default = /*#__PURE__*/_interopDefaultLegacy(Flatbush);
var booleanIntersects__default = /*#__PURE__*/_interopDefaultLegacy(booleanIntersects);
var booleanContains__default = /*#__PURE__*/_interopDefaultLegacy(booleanContains);
var booleanWithin__default = /*#__PURE__*/_interopDefaultLegacy(booleanWithin);
var booleanCrosses__default = /*#__PURE__*/_interopDefaultLegacy(booleanCrosses);
var booleanOverlap__default = /*#__PURE__*/_interopDefaultLegacy(booleanOverlap);
var bbox__default = /*#__PURE__*/_interopDefaultLegacy(bbox);

function merge (geojson1, geojsonOrDataFrame, options = {}) {
  // Parameters
  if (!helpers.isObject(options)) throw new TypeError('options are invalid')
  const mutate = options.mutate || false;
  const onProperty = options.on;

  // validation
  if (!geojson1) throw new Error('<geojson1> is required')
  if (!geojsonOrDataFrame) throw new Error('<geojsonOrDataFrame> is required')
  if (!onProperty) throw new Error('options <on> property is required')
  if (typeof onProperty !== 'string') { throw new TypeError('<onProperty> must be a string (object key)') }
  if (typeof mutate !== 'boolean') { throw new TypeError('<mutate> must be a boolean') }

  // prevent input mutation
  if (mutate === false) {
    try {
      geojson1 = clone__default['default'](geojson1);
    } catch (e) {
      throw new Error('Cloning input GeoJSON failed. Check for missing/empty geometries in the feature collection. Original @turf/clone error message:', e)
    }
  }

  // create an array of all properties of the second GeoJSON input
  // that will be merged into the first
  let propsGeojsonOrDataFrame = [];
  // lazy check if second argument is a geojson object, otherwise just use it straight away
  if (geojsonOrDataFrame.type) {
    meta.featureEach(geojsonOrDataFrame, (currentFeature, featureIndex) => {
      propsGeojsonOrDataFrame.push(currentFeature.properties);
    });
  } else {
    propsGeojsonOrDataFrame = geojsonOrDataFrame;
  }

  // find all features in the second GeoJSON that have the same "onProperty" value
  // as the current feature from the first GeoJSON -> merge these properties into
  // the first GeoJSONs property object
  meta.featureEach(geojson1, currentFeature => {
    const propsToMerge = propsGeojsonOrDataFrame.filter(featureProps => featureProps[onProperty] === currentFeature.properties[onProperty]);

    if (propsToMerge && propsToMerge.length) {
      currentFeature.properties = {
        ...Object.assign(...propsToMerge),
        ...geojson1.properties
      };
    }
  });

  return geojson1
}

function mergeWith (object1, object2, lsuffix, rsuffix) {
  const keys1 = Object.keys(object1);
  const keys2 = Object.keys(object2);
  const allKeys = [...keys1, ...keys2];
  const uniq = [...new Set(allKeys)];
  const feature = {};
  if (allKeys.length === uniq.length) {
    keys1.forEach(key => (feature[key] = object1[key]));
    keys2.forEach(key => (feature[key] = object2[key]));
    return feature
  } else {
    const duplicates = allKeys.filter(key => uniq.indexOf(key) === -1);
    keys1.forEach(key => {
      if (duplicates.indexOf(key) > -1) {
        feature[key + '_' + lsuffix] = object1[key];
      } else {
        feature[key] = object1[key];
      }
    });
    keys2.forEach(key => {
      if (duplicates.indexOf(key) > -1) {
        feature[key + '_' + rsuffix] = object2[key];
      } else {
        feature[key] = object1[key];
      }
    });
    return feature
  }
}

function sjoin (geojson1, geojson2, options) {
  // Optional parameters
  options = options || {};
  if (!helpers.isObject(options)) throw new TypeError('options are invalid')
  const how = options.how || 'inner';
  const op = options.opt || 'intersects';
  const matches = options.matches || 'all';
  const lsuffix = options.lsuffix || 'left';
  const rsuffix = options.rsuffix || 'right';
  const inclLeftIndex = options.inclLeftIndex || false;

  // validation
  if (!geojson1) throw new Error('<geojson1> is required')
  if (!geojson2) throw new Error('<geojson2> is required')
  if (geojson1.type !== 'FeatureCollection') throw new Error('<geojson1> must be a GeoJSON FeatureCollection')
  if (geojson2.type !== 'FeatureCollection') throw new Error('<geojson1> must be a GeoJSON FeatureCollection')
  if (typeof how !== 'string') throw new TypeError('<how> must be a string ("left","right","inner","first")')
  const supportedHows = ['inner', 'left', 'right', 'first'];
  if (supportedHows.indexOf(how) === -1) throw new Error(how, ' is not supported as a <how> operation. Must be any of "left","right","inner","first"')
  if (typeof op !== 'string') { throw new TypeError('<op> must be a string ("intersects","contains","within","crosses","overlaps")') }
  const supportedOps = ['intersects', 'contains', 'within', 'crosses', 'overlaps'];
  if (supportedOps.indexOf(op) === -1) throw new Error(op, ' is not supported as a <op> operation. Must be any of "intersects","contains","within","crosses","overlaps"')
  if (typeof matches !== 'string') { throw new TypeError('<matches> must be a string ("all" or "first")') }
  const supportedMatches = ['first', 'all'];
  if (supportedMatches.indexOf(matches) === -1) throw new Error(matches, ' is not supported as a <matches> operation. Must be any of "first" or "all"')
  if (typeof lsuffix !== 'string') throw new TypeError('<lsuffix> must be a string')
  if (typeof rsuffix !== 'string') throw new TypeError('<rsuffix> must be a string')
  if (typeof inclLeftIndex !== 'boolean') { throw new TypeError('<inclLeftIndex> must be a boolean') }

  let operation;
  switch (op) {
    case 'intersects':
      operation = booleanIntersects__default['default'];
      break
    case 'contains':
      operation = booleanContains__default['default'];
      break
    case 'within':
      operation = booleanWithin__default['default'];
      break
    case 'crosses':
      operation = booleanCrosses__default['default'];
      break
    case 'overlaps':
      operation = booleanOverlap__default['default'];
      break
  }

  // assign left and right GeoJSON dataframes according to the type of join
  // default (inner and left): first GeoJSON is considered as the left DF
  let leftGeojson = geojson1;
  let rightGeojson = geojson2;
  if (how === 'right') {
    leftGeojson = geojson2;
    rightGeojson = geojson1;
  }

  // build spatial index of the right GeoJSON in order to minimize the number of
  // polys/points to check for intersections
  const spatialIndex = new Flatbush__default['default'](rightGeojson.features.length);
  for (const feature of rightGeojson.features) {
    const featureBBOX = bbox__default['default'](feature);
    spatialIndex.add(featureBBOX[0], featureBBOX[1], featureBBOX[2], featureBBOX[3]);
  }
  spatialIndex.finish();

  const results = [];
  const noFeatures = leftGeojson.features.length;

  for (let leftIndex = 0; leftIndex < noFeatures; leftIndex++) {
    if (!leftGeojson.features[leftIndex].geometry) continue
    const leftFeature = leftGeojson.features[leftIndex];
    const featureBBOX = bbox__default['default'](leftFeature);
    const bboxResults = spatialIndex.search(
      featureBBOX[0],
      featureBBOX[1],
      featureBBOX[2],
      featureBBOX[3]
    );

    let matchingRightIndeces = [];
    if (matches === 'first') {
      // in order to speed up the process, we only search for the very first
      // match using Array.find (instead of Array.filter used below)
      const firstMatch = bboxResults
        .find(rightIndex => operation(
          rightGeojson.features[rightIndex],
          leftFeature)
        );
      if (firstMatch) matchingRightIndeces = [firstMatch];
    } else {
      matchingRightIndeces = bboxResults
        .filter(rightIndex => operation(
          rightGeojson.features[rightIndex],
          leftFeature)
        );
    }

    const noMatchingRightIndeces = matchingRightIndeces.length;
    if (noMatchingRightIndeces) {
      for (let i = 0; i < noMatchingRightIndeces; i++) {
        const matchingRightIndex = matchingRightIndeces[i];
        const feature = {
          type: 'Feature',
          geometry: leftFeature.geometry,
          properties: mergeWith(leftFeature.properties, {
            ...(how === 'right' && { index_left: matchingRightIndex }),
            ...(how !== 'right' && { index_right: matchingRightIndex }),
            ...(how === 'right' && inclLeftIndex && { index_right: leftIndex }),
            ...(how !== 'right' && inclLeftIndex && { index_left: leftIndex }),
            ...rightGeojson.features[matchingRightIndex].properties
          })
        };
        results.push(feature);
      }
    } else if (how === 'left' || how === 'right') {
      results.push(leftFeature);
    }
  }

  const fc = {
    type: 'FeatureCollection',
    features: results
  };

  return fc
}

const spatialmerge = {
  merge,
  sjoin
};

module.exports = spatialmerge;
